/*
 *  Copyright (C) 1998,9 by Marco G"otze.
 *
 *  This code is part of the wmpinboard source package, which is
 *  distributed under the terms of the GNU GPL2.
 */

#include <locale.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <stdarg.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#include "wmpinboard.h"
#include "xmisc.h"

extern XpmIcon app;
extern GC NormalGC;
extern Window win, iconwin;
extern Window Root;
extern Pixmap pixmask;

XIC InputContext = 0;
static char *rs_inputMethod = 0;
static char *rs_preeditType = 0;

/*
 * replace o_col by n_col in an XImage ([no]_col are XGetPixel return values!)
 */
void
replace_color(XImage *img, int x, int y, int w, int h, unsigned long o_col,
  unsigned long n_col)
{
  int i, j;

  for (i = x; i < x+w; i++)
    for (j = y; j < y+h; j++)
      if (XGetPixel(img, i, j) == o_col)
        XPutPixel(img, i, j, n_col);
}

/*
 * put dest on src, copying pixels only where src's color is != mask_col's
 */
void
merge_masked(XImage *src, XImage *dest, int sx, int sy, int dx, int dy,
  int width, int height, int mask_col)
{
  unsigned long mc = get_color(mask_col);
  unsigned long p;
  int i, j;

  for (i = 0; i < width; i++)
    for (j = 0; j < height; j++) {
      p = XGetPixel(src, sx+i, sy+j);
      if (p != mc)
        XPutPixel(dest, dx+i, dy+j, p);
    }
}

/*
 * copies a rectangular area within one XImage to another (both images have to
 * be of the SAME FORMAT and characteristics!)
 */
void
copy_image_area(XImage *src, XImage *dest, int sx, int sy, int dx, int dy,
  int w, int h)
{
  int b, y;

  b = src->bits_per_pixel/8;
  for (y = 0; y < h; y++)
    memcpy(&(dest->data[dest->xoffset + dest->bytes_per_line*(dy+y) + dx*b]),
      &(src->data[src->xoffset + src->bytes_per_line*(sy+y) + sx*b]), b*w);
}

/*
 * updates the app's display using a 64x64 area at (x,y) in img
 */
void
update_display(XImage *img, int x, int y)
{
  XPutImage(display, app.pixmap, NormalGC, img, x, y, 0, 0, 64, 64);
  redraw_window();
}

/*
 * converts a pixmap structure to an XpmIcon
 */
void
get_xpm(XpmIcon *app, char *pixmap_bytes[])
{
  XWindowAttributes attributes;
  int err;

  XGetWindowAttributes(display, Root, &attributes);
  app->attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
  err = XpmCreatePixmapFromData(display, Root, pixmap_bytes, &(app->pixmap),
    &(app->mask), &(app->attributes));
	
  if (err != XpmSuccess) {
    fprintf(stderr, "Not enough free colorcells.\n");
    exit(1);
  }
}

/*
 * looks up a color in the global X color map
 */
Pixel
lookup_color(char *name)
{
  XColor color;
  XWindowAttributes attributes;

  XGetWindowAttributes(display, Root, &attributes);
  color.pixel = 0;
  if (!XParseColor(display, attributes.colormap, name, &color)) {
    fprintf(stderr, "Can't parse %s.\n", name);
  } else if (!XAllocColor(display, attributes.colormap, &color)) {
    fprintf(stderr, "Can't allocate %s.\n", name);
  }
  return color.pixel;
}

/*
 * flush Expose events
 */
int
flush_expose(Window w)
{
  XEvent dummy;
  int i=0;

  while (XCheckTypedWindowEvent(display, w, Expose, &dummy)) i++;
  return i;
}

/*
 * redraws the application's window
 */
void
redraw_window(void)
{
  flush_expose(iconwin);
  XCopyArea(display, app.pixmap, iconwin, NormalGC, 0, 0,
    app.attributes.width, app.attributes.height, 0,0);
  flush_expose(win);
  XCopyArea(display, app.pixmap, win, NormalGC, 0, 0,
    app.attributes.width, app.attributes.height, 0,0);
}

/*
 * sets either the app's shape mask for normal operation or reverts to none
 * (edit mode)
 */
void
set_mask(int mask)
{
  static XRectangle rect = { 0, 0, 64, 64 };

  if (mask) {
    XShapeCombineMask(display, win, ShapeBounding, 0, 0, pixmask, ShapeSet);
    XShapeCombineMask(display, iconwin, ShapeBounding, 0, 0, pixmask, ShapeSet);
  } else {
    XShapeCombineRectangles(display, win, ShapeBounding, 0, 0, &rect, 1,
      ShapeSet, 0);
    XShapeCombineRectangles(display, iconwin, ShapeBounding, 0, 0, &rect, 1,
      ShapeSet, 0);
  }
}

/*
 * sets up XIC-orientated keyboard handling
 * [stolen from Eterm 0.8.6, minor adaptations]
 */
void
init_xlocale(void)
{
  char *p, *s, buf[32], tmp[1024];
  XIM xim = 0;
  XIMStyle input_style = 0;
  XIMStyles *xim_styles = 0;
  int found;

  InputContext = 0;
  setlocale(LC_ALL, "");

  if (!rs_inputMethod || !*rs_inputMethod) {
    if ((p = XSetLocaleModifiers("@im=none")) && *p)
      xim = XOpenIM(display, 0, 0, 0);
  } else {
    strcpy(tmp, rs_inputMethod);
    for (s = tmp; *s;) {
      char *end, *next_s;
      
      for (; *s && isspace(*s); s++);
      if (!*s) break;
      end = s;
      for (; *end && (*end != ','); end++);
      next_s = end--;
      for (; (end >= s) && isspace(*end); end--);
      *(end + 1) = '\0';
      
      if (*s) {
        strcpy(buf, "@im=");
        strcat(buf, s);
        if ((p = XSetLocaleModifiers(buf)) && *p
          && (xim = XOpenIM(display, 0, 0, 0)))
        {
          break;
        }
      }
      if (!*next_s) break;
      s = (next_s + 1);
    }
  }

  if (!xim && (p = XSetLocaleModifiers("")) && *p)
    xim = XOpenIM(display, 0, 0, 0);
    
  if (!xim) {
    fprintf(stderr, "Failed to open input method.\n");
    return;
  }

  if (XGetIMValues(xim, XNQueryInputStyle, &xim_styles, 0) || !xim_styles) {
    XCloseIM(xim);
    fprintf(stderr, "Input method doesn't support any style.\n");
    return;
  }

  strcpy(tmp, (rs_preeditType ? rs_preeditType : "Root"));
  for (found = 0, s = tmp; *s && !found;) {
    unsigned short i;
    char *end, *next_s;
  
    while (*s && isspace(*s)) s++;
    if (!*s) break;
    end = s;
    while (*end && (*end != ',')) end++;
    next_s = end--;
    while ((end >= s) && isspace(*end)) *end-- = 0;
  
    if (!strcmp(s, "OverTheSpot"))
      input_style = (XIMPreeditPosition | XIMStatusArea);
    else if (!strcmp(s, "OffTheSpot"))
      input_style = (XIMPreeditArea | XIMStatusArea);
    else if (!strcmp(s, "Root"))
      input_style = (XIMPreeditNothing | XIMStatusNothing);
  
    for (i = 0; i < xim_styles->count_styles; i++) {
      if (input_style == xim_styles->supported_styles[i]) {
        found = 1;
         break;
      }
    }
    s = next_s;
  }
  XFree(xim_styles);
    
  if (found == 0) {
    XCloseIM(xim);
    fprintf(stderr, "Input method doesn't support my preedit type.\n");
    return;
  }

  if (input_style != (XIMPreeditNothing | XIMStatusNothing)) {
    XCloseIM(xim);
    fprintf(stderr, "This program only supports the `Root' preedit type.\n");
    return;
  }

  InputContext = XCreateIC(xim, XNInputStyle, input_style,
    XNClientWindow, win, XNFocusWindow, win, 0);

  if (!InputContext) {
    XCloseIM(xim);
  }
}

